//----------------------------------------------------------------------
//-- AXI stream wrapper for the hard FIFO with arbitrary bit width
//--
//-- The fifo depth is a fixed value based on the width.
//-- 16kb for 32-bit streams, and 32kb for 64-bit streams.
//-- Streams larger than 64-bits use multiple parallel FIFOs.
//----------------------------------------------------------------------

`default_nettype none

module stream_fifo # (
    parameter FIFO_MEMORY_TYPE = "block",
    parameter integer DATA_WIDTH = 32,
    parameter integer FIFO_DEPTH = 512,
    parameter ALMOST_FULL_OFFSET = 13'h0080, //threshold for almost full signal
    parameter ALMOST_EMPTY_OFFSET = 13'h0080, //threshold for almost empty signal
    parameter ENABLE_TKEEP = 0, //set to 1 to use tkeep signals
    parameter PACKET_MODE = 0) //1=block output until complete packet buffered
(
    input wire clk,
    input wire rst,

    input wire [DATA_WIDTH-1:0] in_tdata,
    input wire [(DATA_WIDTH/8)-1:0] in_tkeep,
    input wire in_tlast,
    input wire in_tvalid,
    output wire in_tready,

    output wire [DATA_WIDTH-1:0] out_tdata,
    output wire [(DATA_WIDTH/8)-1:0] out_tkeep,
    output wire out_tlast,
    output wire out_tvalid,
    input wire out_tready,

    output wire almost_empty,
    output wire almost_full
);
    localparam FULL_WIDTH = DATA_WIDTH + (DATA_WIDTH/8);

    wire wr_enabled, rd_enabled;

    wire [DATA_WIDTH-1:0] wfifo_tdata;
    wire [(DATA_WIDTH/8)-1:0] wfifo_tkeep, wfifo_tkeep_i;
    wire wfifo_tlast;
    wire wfifo_tvalid;
    wire wfifo_tready;

    stream_fifo_srl32 #(.DATA_WIDTH(FULL_WIDTH)) in_fifo
    (
        .clk(clk),
        .rst(rst),
        .in_tdata({in_tkeep, in_tdata}),
        .in_tlast(in_tlast),
        .in_tvalid(in_tvalid),
        .in_tready(in_tready),
        .out_tdata({wfifo_tkeep, wfifo_tdata}),
        .out_tlast(wfifo_tlast),
        .out_tvalid(wfifo_tvalid),
        .out_tready(wfifo_tready)
    );

    wire [DATA_WIDTH-1:0] rfifo_tdata;
    wire [(DATA_WIDTH/8)-1:0] rfifo_tkeep, rfifo_tkeep_i;
    wire rfifo_tlast;
    wire rfifo_tvalid;
    wire rfifo_tready;

    wire empty;
    wire full;
    wire wr_rst_busy;
    wire rd_rst_busy;
    wire read = rfifo_tready && !empty && rd_enabled;
    wire write = wfifo_tvalid && !full && wr_enabled;
    assign wfifo_tready = !full && wr_enabled && !wr_rst_busy;
    assign rfifo_tvalid = !empty && rd_enabled && !rd_rst_busy;

    wire [(DATA_WIDTH/8)-1:0] tkeep_zeros = {(DATA_WIDTH/8){1'b0}};

    //encode last into tkeep: 0s on all but last transfer cycles
    //and tkeep cannot be all zero on last, will be inverted
    generate
    if (ENABLE_TKEEP) assign wfifo_tkeep_i = wfifo_tlast?((|(wfifo_tkeep))?wfifo_tkeep:(~tkeep_zeros)):tkeep_zeros;
    else              assign wfifo_tkeep_i = wfifo_tlast?~tkeep_zeros:tkeep_zeros;
    endgenerate

    //decode output tkeep and last
    //last is signaled by a valid tkeep output
    //output tkeep is all 1s until last transfer
    assign rfifo_tlast = |(rfifo_tkeep_i);
    assign rfifo_tkeep = rfifo_tlast?rfifo_tkeep_i:~tkeep_zeros;

    xpm_fifo_sync #(
        .FIFO_MEMORY_TYPE(FIFO_MEMORY_TYPE),
        .FIFO_WRITE_DEPTH(FIFO_DEPTH),
        .WRITE_DATA_WIDTH(FULL_WIDTH),
        .WR_DATA_COUNT_WIDTH($clog2(FIFO_DEPTH)+1),
        .READ_MODE("fwft"),
        .FIFO_READ_LATENCY(0),
        .READ_DATA_WIDTH(FULL_WIDTH),
        .RD_DATA_COUNT_WIDTH($clog2(FIFO_DEPTH)+1),
        .PROG_FULL_THRESH(FIFO_DEPTH-ALMOST_FULL_OFFSET),
        .PROG_EMPTY_THRESH(ALMOST_EMPTY_OFFSET))
    fifo(
        .wr_clk(clk),
        .wr_en(write),
        .din({wfifo_tkeep_i, wfifo_tdata}),
        .full(full),
        .overflow(),
        .wr_rst_busy(wr_rst_busy),
        .sleep(1'b0),
        .rst(rst),
        .rd_en(read),
        .dout({rfifo_tkeep_i, rfifo_tdata}),
        .empty(empty),
        .underflow(),
        .rd_rst_busy(rd_rst_busy),
        .prog_full(almost_full),
        .wr_data_count(),
        .wr_ack(),
        .prog_empty(almost_empty),
        .rd_data_count(),
        .injectsbiterr(1'b0),
        .injectdbiterr(1'b0),
        .sbiterr(),
        .data_valid(),
        .almost_full(),
        .almost_empty(),
        .dbiterr());

    stream_fifo_srl32 #(.DATA_WIDTH(FULL_WIDTH)) out_fifo
    (
        .clk(clk),
        .rst(rst),
        .in_tdata({rfifo_tkeep, rfifo_tdata}),
        .in_tlast(rfifo_tlast),
        .in_tvalid(rfifo_tvalid),
        .in_tready(rfifo_tready),
        .out_tdata({out_tkeep, out_tdata}),
        .out_tlast(out_tlast),
        .out_tvalid(out_tvalid),
        .out_tready(out_tready)
    );

    //maintain count of packets within the stream fifo
    reg [1:0] pkt_count;
    wire wlast = wfifo_tvalid && wfifo_tready && wfifo_tlast;
    wire rlast = rfifo_tvalid && rfifo_tready && rfifo_tlast;
    always @(posedge clk) begin
        if (rst) pkt_count = 2'b0;
        else if (wlast && !rlast) pkt_count = pkt_count + 1'b1;
        else if (!wlast && rlast) pkt_count = pkt_count - 1'b1;
    end

    generate
    if (PACKET_MODE) begin
        //write until one packet is streaming out and other complete one is in the fifo
        assign wr_enabled = pkt_count < 2'h2; //write for packet counts of 0 and 1
        assign rd_enabled = pkt_count != 2'h0; //read when one packet is present
    end
    else begin
        assign wr_enabled = 1'b1;
        assign rd_enabled = 1'b1;
    end
    endgenerate

endmodule //stream_fifo

`default_nettype wire
