////////////////////////////////////////////////////////////////////////////////
// Stream fifo using internally cascaded SRLC32E blocks in a CLB
////////////////////////////////////////////////////////////////////////////////

`default_nettype none

module stream_fifo_srl32 #(
    parameter DATA_WIDTH = 32,
    parameter DEPTH = 1 //1 to 4
) (
    input wire clk,
    input wire rst,

    input wire [DATA_WIDTH-1:0] in_tdata,
    input wire in_tlast,
    input wire in_tvalid,
    output reg in_tready,

    output wire [DATA_WIDTH-1:0] out_tdata,
    output wire out_tlast,
    output reg out_tvalid,
    input wire out_tready
);
    localparam FULL_WIDTH = DATA_WIDTH+1; //DATA + LAST

    //the address range based on the number of SRLC32E
    localparam MAX_ADDR = (DEPTH*32)-2;

    //complete input and output array for each instantiation
    wire [FULL_WIDTH-1:0] D = {in_tlast, in_tdata};
    wire [FULL_WIDTH-1:0] O;
    assign {out_tlast, out_tdata} = O;

    //logic for read and write cycles
    wire write = in_tvalid && in_tready;
    wire read = out_tvalid && out_tready;

    //increment the address when read and write operations occur
    reg [6:0] raddr;
    always @(posedge clk) begin
        if (rst) begin
            raddr <= 7'b0;
            in_tready <= 1'b1;
            out_tvalid <= 1'b0;
        end
        else if (!read && write) begin
            if (raddr == MAX_ADDR) in_tready <= 1'b0;
            if (out_tvalid) raddr <= raddr + 1'b1;
            out_tvalid <= 1'b1;
        end
        else if (read && !write) begin
            if (raddr == 7'b0) out_tvalid <= 1'b0;
            else raddr <= raddr - 1'b1;
            in_tready <= 1'b1;
        end
    end

    //generate cascaded SRLC32E for each bit
    genvar i, j;
    generate
    for (i = 0; i < FULL_WIDTH ; i = i + 1) begin

        wire [3:0] Q;
        wire [4:0] Q31;

        //feed data into the first Q31
        //the rest are between SRL blocks
        assign Q31[0] = D[i];

        //assign output based on upper addr bits
        assign O[i] = Q[raddr[6:5]];

        for (j = 0; j < DEPTH; j = j + 1) begin
            SRLC32E srlc32e_i_j(
                .Q(Q[j]),
                .Q31(Q31[j+1]),
                .A(raddr[4:0]),
                .CE(write),
                .CLK(clk),
                .D(Q31[j])
            );
        end

    end
    endgenerate

endmodule //stream_fifo_srl32

`default_nettype wire
